// ============================================================================
// Project:	Deductive Database
// Filename:	lex.cpp
// Purpose:	Lexical Scanner
// Last Change:	26.08.2015
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 1997-2015 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"

#include <stdlib.h>

#include "err.h"
#include "ch.h"
#include "lex.h"

//=============================================================================
// Exported Functions:
//=============================================================================

//-----------------------------------------------------------------------------
// Constructor:
//-----------------------------------------------------------------------------

lex_c::lex_c(in_c *input)
{
	// Check Parameter:
	CHECK(input->is_open(), "lex_c::constructor: Input stream not open");

	// Initialize data members:
#ifdef	VER_DEBUG
	Magic	= LEX_MAGIC;
#endif
	Input	= input;
	Eof	= false;
	Buf[0]	= '\n';
	Buf[1]	= '\0';
	Bytes	= 1;
	LineNo	= 0;
	LinePtr = Buf;
	LineEnd = Buf;
	TokType	= TOK_EOF;
	TokPtr	= Buf;
	TokEnd	= Buf;
	TokInt	= 0;

	// Read first token:
	next();
}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

lex_c::~lex_c(void)
{
	// Invalidate object:
#ifdef	VER_DEBUG
	Magic = 0;
#endif
}

//-----------------------------------------------------------------------------
// Integrity check (only for debugging purposes):
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG

str_t	lex_c::check() const
{
	if(Magic != LEX_MAGIC)
		return "invalid lex object (bad magic number)";
	if(Bytes < 0)
		return "invalid lex object (bytes in buffer < 0)";
	if(Bytes > LEX_LINESIZE+LEX_BLOCKSIZE)
		return "invalid lex object (bytes in buffer > buffer size)";
	if(Buf[Bytes] != '\0')
		return "invalid lex object (buffer not null-terminated)";
	if(LineNo < 0)
		return "invalid lex object (bad line number)";
	if(LinePtr < Buf)
		return "invalid lex object (line ptr < buffer)";
	if(LinePtr > LineEnd)
		return "invalid lex object (line end < line begin)";
	if(!(LineEnd < Buf+Bytes || (Eof && LineEnd==Buf+Bytes)))
		return "invalid lex object (line end > buffer end)";
	if(!tok_valid(TokType))
		return "invalid lex object (bad token type)";
	if(TokPtr < LinePtr)
		return "invalid lex object (begin of token < begin of line)";
	if(TokPtr > TokEnd)
		return "invalid lex object (token begin > token end)";
	if(TokEnd > LineEnd)
		return "invalid lex object (token end > line end)";

	// No error found:
	return 0;
}

#endif

//-----------------------------------------------------------------------------
// getline: Fill input buffer.
//-----------------------------------------------------------------------------

void	lex_c::getline(void)
{
	register char	*p_end;
	register char	*p_from;
	int		bytes;

	// Integrity check (No full check: Token pointers may be invalid!)
	CHECK(!Eof, "lex_c::getline(at eof)");

	// Update counters, pointers:
	p_end = LineEnd + 1;
	LinePtr = p_end;
	LineNo++;

	// Find next line end:
	while(*p_end && !ch_c::is_newline(*p_end))
		p_end++;
	if(*p_end) {
		LineEnd = p_end;
		return;
	}

	// Ok, line is not complete. Is it really at the end of the buffer?
	if(p_end - Buf != Bytes){
		err_c::null_char(Input->filename(), LineNo);
		exit(1);
	}

	// Now shift current line to start of buffer :
	p_end = Buf;
	p_from = LinePtr;
	LinePtr = p_end;
	Bytes -= (p_from - p_end);
	while(*p_from)
		*p_end++ = *p_from++;
	*p_end = '\0';

	// Read input until line is complete:
	while(*p_end == '\0' && !Eof) {
		// Is now enough room in the buffer?
		if(Bytes >= LEX_LINESIZE) {
			err_c::line_too_long(Input->filename(), LineNo);
			exit(1);
		}

		// Now read from the file:
		bytes = Input->read(p_end, LEX_BLOCKSIZE);

		// Did an error occur?
		CHECK(bytes<=LEX_BLOCKSIZE, "lex_c::getline(read too much)");
		if(bytes < 0) {
			// Input->read() already printed an error message.
			bytes = 0;
		}
		Eof = (bytes==0);

		// Terminate buffer contents:
		Bytes += bytes;
		Buf[Bytes] = '\0';

		// Now continue to find the line end:
		while(*p_end && !ch_c::is_newline(*p_end))
			p_end++;
	}

	// Ok, line end found:
	LineEnd = p_end;
}



//-----------------------------------------------------------------------------
// next: Read next token.
//-----------------------------------------------------------------------------

void	lex_c::next()
{
	register char *p;

	// File must be open:
	CHECK_VALID("lex_c::next");
	CHECK(Input->is_open(), "lex_c::next: File not open");

	// Use register variable for pointer to next character:
	p = TokEnd;

	// Skip white space and comments (strange, but efficient loop):
	for(;;) {
		if(ch_c::is_space(*p))
			p++;
		else if(ch_c::is_newline(*p)) {
			CHECK(p==LineEnd, "lex_c::next: bad eol");
			getline();
			p = LinePtr;
			continue;
		}
		else if(ch_c::is_comment(*p))
			p = LineEnd;
		else
			break;
	}

	// Remember position of token:
	TokPtr = p;

#ifdef VER_DEBUG
	// Reset integer value:
	TokInt = 0;
#endif

	// Determine token type:
	switch(ch_c::classify(*p)) {
		case CH_NULL:
			TokType = TOK_EOF;
#ifdef LEX_INTERACTIVE_SUPPORT
			if(Interactive) {
				*p = CH_NEWLINE;
				*(p + 1) = '\0';
				Eof = false;
				Bytes++;
			}
#endif
			break;
		case CH_EOF:
			TokType = TOK_EOF;
			break;
		case CH_ATOM:
			p++;
			while(ch_c::is_alnum(*p))
				p++;
			if(p-TokPtr == 3 && TokPtr[0] == 'n'
				&& TokPtr[1] == 'o' && TokPtr[2] == 't') {
				TokType = TOK_NOT;
			}
			else
				TokType = TOK_ATOM;
			break;
		case CH_VAR:
			p++;
			while(ch_c::is_alnum(*p))
				p++;
			if(ch_c::is_undersc(*TokPtr) && p == TokPtr + 1)
				TokType = TOK_ANONVAR;
			else
				TokType = TOK_VAR;
			break;
		case CH_MINUS:
			p++;
			if(ch_c::classify(*p)==CH_DIGIT) {
				p++;
				while(ch_c::classify(*p)==CH_DIGIT)
					p++;
				TokType = TOK_INT;
			}
			else
				TokType = TOK_MINUS;
			break;
		case CH_DIGIT:
			p++;
			while(ch_c::classify(*p)==CH_DIGIT)
				p++;
			TokType = TOK_INT;
			break;
		case CH_QUOTE:
			p++;
			while(*p && !ch_c::is_quote(*p)
				&& !ch_c::is_newline(*p)) {
				if(ch_c::is_backslash(*p)) {
					p++;
					if(!ch_c::is_escape_seq(*p)) {
						p--;
						break;
					}
				}
				p++;
			}
			if(ch_c::is_backslash(*p))
				TokType = TOK_BAD_ESCAPE_SEQ;
			else if(!ch_c::is_quote(*p))
				TokType = TOK_MISSINGQ;
			else {
				TokType = TOK_STR;
				p++;
			}
			break;
		case CH_FULLSTOP:
			p++;
			if(ch_c::is_space(*p)||ch_c::is_newline(*p)||*p=='\0')
				TokType = TOK_FULLSTOP;
			else
				TokType = TOK_NOSPACE;
			break;
		case CH_COMMA:
			p++;
			TokType = TOK_COMMA;
			break;
		case CH_LPAREN:
			p++;
			TokType = TOK_LPAREN;
			break;
		case CH_RPAREN:
			p++;
			TokType = TOK_RPAREN;
			break;
		case CH_LESS:
			p++;
			if(ch_c::classify(*p) == CH_MINUS) {
				p++;
				TokType = TOK_IF;
			}
			else
				TokType = TOK_LESS;
			break;
		case CH_GREATER:
			p++;
			TokType = TOK_GREATER;
			break;
		case CH_COLON:
			p++;
			if(ch_c::classify(*p) == CH_MINUS) {
				p++;
				TokType = TOK_IF;
			}
			else
				TokType = TOK_COLON;
			break;
		case CH_NOT:
			p++;
			TokType = TOK_NOT;
			break;
		case CH_OR:
		case CH_SEMICOLON:
			p++;
			TokType = TOK_OR;
			break;
		case CH_AND:
			p++;
			TokType = TOK_AND;
			break;
		case CH_QUEST:
			p++;
			TokType = TOK_QUERY;
			break;
		case CH_UNUSED:
		case CH_UNDEFINED:
			// The parser will print an error message ...
			p++;
			TokType = TOK_ILLEGAL_CHAR;
			break;
		default:
			CHECK_IMPOSSIBLE("lex::next(bad character class)");
	}

	// Remember pointer position:
	TokEnd = p;

	// Convert integer:
	if(TokType == TOK_INT) {
		if(!scan_int(TokPtr, TokEnd, &TokInt))
			TokType = TOK_INT_OVERFLOW;
	}
}

//-----------------------------------------------------------------------------
// show_err: Show syntax error for the current token in the current line.
//-----------------------------------------------------------------------------

void	lex_c::show_err(syn_t syn) const
{
	register char *p;

	// File must be open:
	CHECK_VALID("lex_c::show");
	CHECK(Input->is_open(), "lex_c::show: File not open");

	// If the current token is a "lexical error", that is the problem:
	switch(TokType) {
		case TOK_ILLEGAL_CHAR:
			syn = SYN_ILLEGAL_CHAR;
			break;
		case TOK_MISSINGQ:
			syn = SYN_MISSINGQ;
			break;
		case TOK_NOSPACE:
			syn = SYN_NOSPACE;
			break;
		case TOK_INT_OVERFLOW:
			syn = SYN_INT_OVERFLOW;
			break;
		case TOK_BAD_ESCAPE_SEQ:
			syn = SYN_BAD_ESCAPE_SEQ;
			break;
		default:
			break;
	}

	// Print line:
	err_c::syntax_begin(Input->filename(), LineNo, syn_text(syn));
	for(p = LinePtr; *p && *p != '\n'; p++)
		err_c::syntax_input(*p);
	err_c::syntax_newline();

	// Show position:
	for(p = LinePtr; *p && *p != '\n' && p < TokPtr; p++)
		err_c::syntax_input((*p == '\t') ? '\t' : ' ');
	err_c::syntax_input('^');
	if(*TokPtr == '\0')
		err_c::syntax_eof();
	err_c::syntax_end();
}

//-----------------------------------------------------------------------------
// scan_int: Convert current token into integer.
//-----------------------------------------------------------------------------

bool	lex_c::scan_int(register str_t p, register str_t end_p, int *value)
{
	int    n;
	int digit_val;
	bool neg;

	// Check parameters:
	CHECK(p != 0, "lex_c::scan_int: p is NIL");
	CHECK(end_p != 0, "lex_c::scan_int: end_p is NIL");
	CHECK(value != 0, "lex_c::scan_int: value is NIL");
	CHECK(p < end_p, "lex_c::scan_int: p >= end_p");

	// Initialize conversion result:
	n = 0;

	// Check for minus sign:
	if(*p == '-') {
		neg = true;
		p++;
	}
	else
		neg = false;

	// Loop over digits:
	while(p < end_p) {
		switch(*p) {
			// We do the computation with negative numbers,
			// because there is usually one negative number
			// more than positive numbers.
			case '0': digit_val =  0; break;
			case '1': digit_val = -1; break;
			case '2': digit_val = -2; break;
			case '3': digit_val = -3; break;
			case '4': digit_val = -4; break;
			case '5': digit_val = -5; break;
			case '6': digit_val = -6; break;
			case '7': digit_val = -7; break;
			case '8': digit_val = -8; break;
			case '9': digit_val = -9; break;
			default:  digit_val = 0;
			CHECK_IMPOSSIBLE("lex_c::scan_int: Illegal character");
		}
		// Overflow check (without generating an overflow).
		// Remember that n, digit_val, MAX_NEG_INT are all negative.
		// We have to check that the new value of n is still within
		// the limits, i.e. there will be an overflow (underflow) if:
		// (10 * n) + digit_val < LEX_MIN_INT.
		// This gives the following condition for n:
		if(n < (LEX_MIN_INT-digit_val)/10) {
			*value = 0;
			return false;
		}

		// Compute new value of n:
		n = 10 * n + digit_val;

		// Ok, this character is processed:
		p++;
	}

	// If it is not a negative number, make it positive:
	if(!neg) {
		// Check for overflow:
		if(n < -LEX_MAX_INT) {
			*value = 0;
			return false;
		}
		n = -n;
	}

	// Assign value, return "success":
	*value = n;
	return true;
}
